import pytest
from mock import Mock

from orchestrate.common import TemporaryDirectory
from orchestrate.core.cluster.object import AWSCluster
from orchestrate.core.cluster_metadata.errors import *
from orchestrate.core.cluster_metadata.service import ClusterMetadataService
from orchestrate.core.custom_cluster.service import CustomClusterService
from orchestrate.core.provider.constants import Provider
from orchestrate.core.provider.broker import ProviderBroker

# pylint: disable=protected-access

class TestClusterService(object):
  @pytest.fixture
  def services(self):
    mock_services = Mock(
      aws_service=Mock(
        create_cluster_object=lambda cluster_name: AWSCluster(dict(name=cluster_name)),
      ),
    )
    mock_services.custom_cluster_service = CustomClusterService(mock_services)
    mock_services.provider_broker = ProviderBroker(mock_services)
    return mock_services

  @pytest.fixture
  def cluster_metadata_service(self, services):
    cluster_metadata_service = ClusterMetadataService(services)
    services.cluster_metadata_service = cluster_metadata_service
    return cluster_metadata_service

  @pytest.mark.parametrize('provider', [
    Provider.AWS,
    Provider.CUSTOM,
  ])
  def test_custom_cluster(self, cluster_metadata_service, provider):
    with TemporaryDirectory() as root_dirname:
      cluster_metadata_service._metadata_dir = root_dirname

      provider_service = cluster_metadata_service.services.provider_broker.get_provider_service(provider)
      cluster = provider_service.create_cluster_object(cluster_name='foobar')
      cluster_metadata_service.write_metadata(cluster)

      cluster = cluster_metadata_service.read_metadata('foobar')
      assert cluster.name == 'foobar'
      assert cluster.provider == provider

  def test_double_write(self, cluster_metadata_service):
    with TemporaryDirectory() as root_dirname:
      cluster_metadata_service._metadata_dir = root_dirname

      cluster = cluster_metadata_service.services.custom_cluster_service.create_cluster_object(cluster_name='foobar')
      cluster_metadata_service.write_metadata(cluster)

      with pytest.raises(MetadataAlreadyExistsError):
        cluster_metadata_service.write_metadata(cluster)

  def test_no_metadata(self, cluster_metadata_service):
    with TemporaryDirectory() as root_dirname:
      cluster_metadata_service._metadata_dir = root_dirname

      with pytest.raises(Exception):
        cluster_metadata_service.read_metadata('foobar')
